from enum import Enum
from typing import Tuple
import numpy as np
from sadie.agents import AbstractAgent
from sadie.agents.exceptions import NoTargetError
from sadie.agents.mixins import TargetingMixin


class AgentStates(Enum):
    """
    Enumeration of possible agent states.
    """
    HALTED = 0
    MOVING = 1
    WAITING = 2


class SpatialAgent(AbstractAgent):
    """
    The `SpatialAgent` class creates spatially enabled agents in Euclidean space. These objects are located in space,
    endowed with an `x` and `y` coordinate. These can be set directly.
    """

    def __init__(self, x_init: float, y_init: float, *args, **kwargs):
        super(SpatialAgent, self).__init__()
        self._x = x_init
        self._y = y_init

    @property
    def x(self):
        """
        Returns the x coordinate of the agent. Equivalent to `self.position()[0]`.
        """
        return self._x

    @x.setter
    def x(self, value):
        raise ValueError("Cannot set x coordinate directly. Use either a movement method or set the position using the "
                         "`position` setter.")

    @property
    def y(self):
        """
        Returns the y coordinate of the agent. Equivalent to `self.position()[1]`.
        """
        return self._y

    @y.setter
    def y(self, value):
        raise ValueError("Cannot set y coordinate directly. Use either a movement method or set the position using the "
                         "`position` setter.")

    @property
    def position(self) -> Tuple[float, float]:
        """
        Returns the position vector of the agent as a tuple :math:`(x, y)`.
        """
        return self.x, self.y

    @position.setter
    def position(self, pos: Tuple[float, float]):
        """
        Sets agent position to the tuple `pos`.

        :param pos: agent position
        """
        self._x, self._y = pos

    def distance_from(self, point: Tuple[float, float]) -> float:
        """
        Returns the agent's distance from an arbitrary point `point` defined as a tuple.

        :param point: point to which distance is calculated
        :return: distance between the agent's position and the provided point
        """
        return np.sqrt((self.x - point[0])**2 + (self.y - point[1]) ** 2)

    def vector_to(self, point: Tuple[float, float]) -> Tuple[float, float]:
        """
        Returns the angle and magnitude :math:`(\\theta, r)` to the arbitrary point `point`.

        :param point: point to which vector is calculated
        :return: polar form vector `(theta, r)` between the agent's current location and `point`
        """
        return np.arctan2(point[1] - self.y, point[0] - self.x), self.distance_from(point)

    def update(self):
        pass

    def report(self):
        pass


class MovingSpatialAgent(SpatialAgent):
    """
    A `MovingSpatialAgent` is a `SpatialAgent` that has movement abilities, in particular:

    * `move_to(x, y)`: move to an arbitrary point :math:`(x, y)`
    * `move_by(x, y)`: move by a vector :math:`(x, y)`, and
    * `move_p(theta, r)`: move by a vector :math:`(\\theta, r)`
    """
    def __init__(self, x_init: float, y_init: float, *args, **kwargs):
        super(MovingSpatialAgent, self).__init__(x_init=x_init, y_init=y_init)
        self._distance_traveled: float = 0.0
        self._trip_distance: float = 0.0

    @property
    def distance_traveled(self):
        return self._distance_traveled

    @distance_traveled.setter
    def distance_traveled(self, value):
        self._distance_traveled = value
        self._trip_distance = value

    @property
    def trip_distance(self):
        return self._trip_distance

    @trip_distance.setter
    def trip_distance(self, value):
        raise ValueError("Cannot set trip distance directly.")

    def reset_trip_odometer(self):
        self._trip_distance = 0.0

    def move_to(self, x: float, y: float):
        """
        Moves the agent to the coordinates :math:`(x, y)`.

        :param x: destination x coordinate
        :param y: destination y coordinate
        """
        self.distance_traveled += np.sqrt((x - self.x) ** 2 + (y - self.y) ** 2)
        self.position = (x, y)

    def move_by(self, x: float, y: float):
        """
        Moves the agent by the vector :math:`(x, y)`, i.e. performs a vector addition to the current position vector.

        :param x: x increment to move by
        :param y: y increment to move by
        """
        self.distance_traveled += np.sqrt(x ** 2 + y ** 2)
        self.position = (self.x + x, self.y + y)

    def move_p(self, theta: float, r: float):
        """
        Moves the agent by a vector defined by the polar coordinates :math:`(\\theta, r)`.

        :param theta: angular component :math:`\\theta`
        :param r: magnitude :math:`r`
        """
        self.distance_traveled += r
        self.position = (self.x + np.cos(theta) * r, self.y + np.sin(theta) * r)


class TargetableAgent(MovingSpatialAgent, TargetingMixin):
    """
    A `TargetableAgent` is a `MovingSpatialAgent` that has, in addition to its position, a target, which it pursues
    every time it is updated.

    In general, you should use the concrete classes derived from this class rather than use it directly, as its update
    class is not defined.
    """
    def __init__(self, x_init: float, y_init: float, velocity: float = 1.0):
        super(TargetableAgent, self).__init__(x_init=x_init, y_init=y_init)
        self._velocity = velocity
        self._state = AgentStates.HALTED

    @property
    def state(self) -> AgentStates:
        """
        Returns the agent's current state.

        :return: agent state
        """
        return self._state

    @state.setter
    def state(self, value: AgentStates):
        self._state = value

    @property
    def velocity(self) -> float:
        """
        Returns the agent's set velocity.

        :return: the agent's velocity
        """
        return self._velocity

    @velocity.setter
    def velocity(self, value: float):
        self._velocity = value

    def move(self):
        """
        Moves agent along to its target.
        """
        if self.target == (None, None):
            raise NoTargetError
        else:
            self.state = AgentStates.MOVING
            if self.target_distance < self.velocity:
                self.distance_traveled += self.target_distance
                self.move_to(*self.target)
            else:
                self.move_p(self.target_azimuth, self.velocity)

    def wait(self):
        """
        Waiting method, implementing a time step during which the agent is not changing state and does not move.
        """
        self.state = AgentStates.WAITING

    def report(self) -> dict:
        """
        Returns a dictionary of agent properties, implementing the reporting method to the model.

        :return: agent property dictionary
        """
        return {"x": self.x,
                "y": self.y,
                "tx": self.target[0],
                "ty": self.target[1],
                "ta": self.target_azimuth if self.target[0] else None,
                "tr": self.target_distance if self.target[0] else None,
                "d": self.distance_traveled,
                "st": self.state}

    def update(self):
        pass



from typing import Optional, Callable
from scipy import stats

from sadie.agents.exceptions import NoTargetError
from sadie.agents.spatial import TargetableAgent, AgentStates
import numpy as np
from numpy import pi as π


class BaseWalker(TargetableAgent):
    """
    The `BaseWalker` is a base class that implements the basic notion of a _walker_: an agent that sets and follows
    a pattern of sequentially assigned targets autonomously.

    The `BaseWalker` does not implement waiting (i.e. it's always on the move) and it does not exhibit extraneously
    sensitive behaviour (i.e. it does not react to the space it is moving across), and has a constant velocity.
    """
    def retarget(self):
        self.set_polar_target(np.random.uniform(0, 2*π), np.random.randint(1, 100))

    def update(self):
        if self.target == (None, None):
            self.retarget()
        elif self.is_on_target:
            self.retarget()
        else:
            self.move()


class WaitingUniformWalker(BaseWalker):
    def __init__(self, x_init: float, y_init: float, velocity: float = 1.0, wait_transition_probability: float = 0.4):
        super(WaitingUniformWalker, self).__init__(x_init=x_init, y_init=y_init, velocity=velocity)
        self.wait_transition_probability = wait_transition_probability

    def update(self):
        if self.target == (None, None):
            self.retarget()
        elif self.is_on_target:
            if np.random.random() > self.wait_transition_probability:
                self.retarget()
            else:
                self.wait()
        else:
            self.move()


class UniformLevyRandomWalker(BaseWalker):
    """
    The Uniform-Lévy random walker determines its trip distances according to the Lévy distribution

    .. math::

        f(x) =  \\frac{e^{- \\frac{1}{2x}}}{\\sqrt{2\\pi x^3}}

    and the azimuth of each target is drawn from a uniform distribution of :math:`(0, 2\\pi]`.
    """

    def retarget(self):
        self.set_polar_target(np.random.uniform(0, 2*π), stats.levy().rvs())


class BoundedUniformLevyRandomWalker(BaseWalker):
    def __init__(self,
                 x_init: float,
                 y_init: float,
                 velocity: float = 1.0,
                 wait_transition_probability: Optional[float] = None,
                 bounding_distribution: stats.rv_continuous = stats.norm,
                 **kwargs):
        super(BoundedUniformLevyRandomWalker, self).__init__(x_init=x_init, y_init=y_init, velocity=velocity)
        self.wait_transition_probability = wait_transition_probability
        self.bounding_distribution = bounding_distribution
        self.kwargs = kwargs

    def update(self):
        if self.target == (None, None):
            self.retarget()
        elif self.is_on_target:
            self.retarget()
        elif self.trip_distance >= self.bounding_distribution.rvs(**self.kwargs):
            self.retarget()
            self.reset_trip_odometer()
        else:
            self.move()


# class UniformRandomWalkAgent(TargetableAgent):
#     def __init__(self, x_init: float, y_init: float, velocity=1, epsilon=1e-4,
#                  r_min: float = 1, r_max: float = 10):
#         super(UniformRandomWalkAgent, self).__init__(x_init=x_init,
#                                                      y_init=y_init,
#                                                      velocity=velocity,
#                                                      epsilon=epsilon)
#         self.r_min, self.r_max = r_min, r_max
#
#     def retarget(self):
#         theta, r = np.random.uniform(0, 2 * π), np.random.uniform(self.r_min, self.r_max)
#         self.set_polar_target(theta, r)
#
#
# class UniformLevyRandomWalkAgent(TargetableAgent):
#     def __init__(self, x_init: float, y_init: float, velocity=1, epsilon=1e-4):
#         super(UniformLevyRandomWalkAgent, self).__init__(x_init=x_init,
#                                                          y_init=y_init,
#                                                          velocity=velocity,
#                                                          epsilon=epsilon)
#
#     def retarget(self):
#         theta, r = np.random.uniform(0, 2 * π), stats.levy().rvs()
#         self.set_polar_target(theta, r)
#
#
# class BoundedDistanceLevyRandomWalkAgent(UniformLevyRandomWalkAgent):
#     def __init__(self, x_init: float, y_init: float, velocity=1, epsilon=1e-4,
#                  mu: float = 100, sigma: float = 12.5):
#         super(UniformLevyRandomWalkAgent, self).__init__(x_init=x_init,
#                                                          y_init=y_init,
#                                                          velocity=velocity,
#                                                          epsilon=epsilon)
#         self.mu, self.sigma = mu, sigma
#
#     def update(self):
#         if self.is_targeted:
#             if self.on_target:
#                 self.retarget()
#                 self.move()
#             else:
#                 if self.distance_traveled > stats.norm(self.mu, self.sigma).rvs():
#                     self.retarget()
#                     self.distance_traveled = 0
#                 self.move()
#         else:
#             self.retarget()
#
#
# class BoundedDistanceLevyRandomWalkAgentWithWait(BoundedDistanceLevyRandomWalkAgent):
#     def __init__(self,
#                  x_init: float,
#                  y_init: float,
#                  velocity=1, epsilon=1e-4,
#                  mu: float = 100,
#                  sigma: float = 12.5,
#                  rest_transition_probability: float = 0.8):
#
#         super(UniformLevyRandomWalkAgent, self).__init__(x_init=x_init,
#                                                          y_init=y_init,
#                                                          velocity=velocity,
#                                                          epsilon=epsilon)
#         self.mu, self.sigma = mu, sigma
#         self.rtp = rest_transition_probability
#
#     def wait(self):
#         if stats.uniform(0, 1).rvs() > self.rtp:
#             self.retarget()
#
#     def update(self):
#         if self.is_targeted:
#             if self.on_target:
#                 self.wait()
#             else:
#                 if self.distance_traveled > stats.norm(self.mu, self.sigma).rvs():
#                     self._target = [self.x, self.y]
#                     self.distance_traveled = 0
#                     self.wait()
#                 else:
#                     self.move()
#         else:
#             self.retarget()
#
#
# class EnergyDependentLevyAgent(BoundedDistanceLevyRandomWalkAgentWithWait):
#     def __init__(self,
#                  x_init: float,
#                  y_init: float,
#                  velocity=1, epsilon=1e-4,
#                  mu: float = 100,
#                  sigma: float = 12.5,
#                  mu_energy: float = 100,
#                  sigma_energy: float = 20,
#                  replenishment_rate: float = 10):
#
#         super(UniformLevyRandomWalkAgent, self).__init__(x_init=x_init,
#                                                          y_init=y_init,
#                                                          velocity=velocity,
#                                                          epsilon=epsilon)
#         self.mu, self.sigma = mu, sigma
#         self.maximum_energy = stats.norm(self.mu, self.sigma).rvs()
#         self.replenishment_rate = replenishment_rate
#         self.energy = self.maximum_energy
#         self.state = "halted"
#
#     def wait(self):
#         self.energy += self.maximum_energy / self.replenishment_rate
#
#     def move(self):
#         if self.tgt_distance > self.velocity:
#             self.x += np.cos(self.heading) * self.velocity
#             self.y += np.sin(self.heading) * self.velocity
#             self.distance_traveled += self.velocity
#             self.energy -= self.velocity
#         else:
#             self.distance_traveled += self.tgt_distance
#             self.energy -= self.tgt_distance
#             self.x, self.y = self._target
#
#     def update(self):
#         if self.state == "halted":
#             self.retarget()
#             self.move()
#             self.state = "moving"
#         elif self.is_targeted:
#             if self.on_target:
#                 self.wait()
#                 if self.energy >= 0.4 * self.maximum_energy:
#                     self.retarget()
#             else:
#                 if self.distance_traveled > stats.norm(self.mu, self.sigma).rvs():
#                     self._target = [self.x, self.y]
#                     self.distance_traveled = 0
#                     self.wait()
#                     self.state = "waiting"
#                     self.retarget()
#                 elif self.state == "waiting":
#                     if self.energy <= self.maximum_energy * 0.8:
#                         self.wait()
#                     else:
#                         self.state = "moving"
#                         self.move()
#                 else:
#                     self.state = "moving"
#                     self.move()
#         else:
#             self.retarget()
#
#     def report(self):
#         return {"id": self.id,
#                 "x": self.x,
#                 "y": self.y,
#                 "energy": self.energy,
#                 "efrac": self.energy/self.maximum_energy,
#                 "state": self.state,
#                 "d": self.tgt_distance,
#                 "trip_length": self.distance_traveled}



import unittest
from random import randint
import numpy as np
from numpy import pi as π
from sadie.agents.spatial import TargetableAgent, AgentStates
from sadie.agents.exceptions import NoTargetError


class TestTargetable(unittest.TestCase):
    def test_state(self):
        a = TargetableAgent(x_init=randint(-100, 100), y_init=randint(-100, 100))
        self.assertEqual(a.state, AgentStates.HALTED, msg="The initial agent state should be HALTED.")
        a.set_absolute_target(x=randint(-100, 100), y=randint(-100, 100))
        a.move()
        self.assertEqual(a.state, AgentStates.MOVING, msg="Following a move, the agent state should be MOVING.")
        a.wait()
        self.assertEqual(a.state, AgentStates.WAITING, msg="Following wait(), the agent state should be WAITING.")

    def test_wait_behaviour(self):
        a = TargetableAgent(x_init=randint(-100, 100), y_init=randint(-100, 100))
        pos = (a.x, a.y)
        a.wait()
        self.assertEqual(pos[0], a.x, msg="The expected behaviour for wait() should not change the agent's position.")
        self.assertEqual(pos[1], a.y, msg="The expected behaviour for wait() should not change the agent's position.")

    def test_move(self):
        a = TargetableAgent(0, 0)
        a.set_absolute_target(1, 0)
        a.move()
        self.assertEqual(a.x, 1)
        self.assertEqual(a.y, 0)
        self.assertEqual(a.distance_traveled, 1)

    def test_velocity_dependent_distance_reach(self):
        a = TargetableAgent(0, 0, velocity=0.5)
        a.set_polar_target(0, 2)
        self.assertFalse(a.is_on_target)
        a.move()
        self.assertFalse(a.is_on_target)
        a.move()
        self.assertFalse(a.is_on_target)
        a.move()
        self.assertFalse(a.is_on_target)
        a.move()
        self.assertTrue(a.is_on_target)

    def test_target_error(self):
        a = TargetableAgent(0, 0)
        with self.assertRaises(NoTargetError, msg="Moving an untargeted TargetableAgent should raise a NoTargetError."):
            a.move()

        with self.assertRaises(NoTargetError, msg="An untargeted TargetableAgent's target_distance property should "
                                                  "raise a NoTargetError."):
            a.target_distance()

        with self.assertRaises(NoTargetError, msg="An untargeted TargetableAgent's target_azimuth property should "
                                                  "raise a NoTargetError."):
            a.target_azimuth()

        with self.assertRaises(NoTargetError, msg="An untargeted TargetableAgent's is_on_target property should raise "
                                                  "a NoTargetError."):
            a.is_on_target()

    def test_report(self):
        ax, ay, dx, dy, tx, ty = np.random.uniform(-100, 100, 6)
        a = TargetableAgent(ax, ay)

        # In general, an untargeted TargetableAgent should return x and y position, 0 distance and None to azimuth and
        # distance to target, as well as None for target coordinates. Its state should be HALTED.
        self.assertEqual(a.report()["x"], ax, msg="At start, the agent should report the correct x position.")
        self.assertEqual(a.report()["y"], ay, msg="At start, the agent should report the correct y position.")
        self.assertIsNone(a.report()["tx"], msg="At start, the target x coordinate should be None (undefined).")
        self.assertIsNone(a.report()["ty"], msg="At start, the target y coordinate should be None (undefined).")
        self.assertIsNone(a.report()["ta"], msg="At start, the target azimuth should be None (undefined).")
        self.assertIsNone(a.report()["tr"], msg="At start, the target distance should be None (undefined).")
        self.assertEqual(a.report()["d"], 0, msg="At start, the agent should have a distance travelled of zero.")
        self.assertEqual(a.report()["st"], AgentStates.HALTED, msg="At start, the agent should be in HALTED state.")

        theta, r = np.random.uniform(0, 2*π), np.random.randint(2, 100)
        a.set_polar_target(theta, r)

        # Now that the target has been set, we expect it to show up in the targeting report.
        self.assertAlmostEqual(a.report()["x"], ax, msg="Targeting should not change the x coordinate of the agent.")
        self.assertAlmostEqual(a.report()["y"], ay, msg="Targeting should not change the x coordinate of the agent.")
        self.assertAlmostEqual(a.report()["tx"], ax + r * np.cos(theta))
        self.assertAlmostEqual(a.report()["ty"], ay + r * np.sin(theta))
        self.assertAlmostEqual(a.report()["ta"], theta, msg=f"Angle for a polar angle movement of {theta},{r} should be"
                                                            f" {theta}.")
        self.assertAlmostEqual(a.report()["tr"], r, msg=f"Distance for a polar angle movement of {theta},{r} should be "
                                                        f"{r}.")
        self.assertEqual(a.report()["d"], 0, msg="Targeting should not change the distance travelled of the agent.")
        self.assertEqual(a.report()["st"], AgentStates.HALTED, msg="After targeting, the agent should still be in a "
                                                                   "HALTED state.")

        a.move()

        # We have now moved the object. Our report should now have the same target, but at an advanced location, in a
        # moving state and having recorded a distance.
        self.assertAlmostEqual(a.report()["x"], ax + np.cos(theta), msg=f"After the first movement, the agent's x "
                                                                  f"coordinate should be {np.cos(theta)} away from "
                                                                  f"its origin (i.e. at {ax + np.cos(theta)} given an "
                                                                  f"origin of {ax}.")
        self.assertAlmostEqual(a.report()["y"], ay + np.sin(theta), msg=f"After the first movement, the agent's y "
                                                                  f"coordinate should be {np.sin(theta)} away from "
                                                                  f"its origin (i.e. at {ay + np.sin(theta)} given an "
                                                                  f"origin of {ay}.")
        self.assertAlmostEqual(a.report()["tx"], ax + r * np.cos(theta), msg="After the first movement, the target x "
                                                                             "coordinate should still be the same.")
        self.assertAlmostEqual(a.report()["ty"], ay + r * np.sin(theta), msg="After the first movement, the target y "
                                                                             "coordinate should still be the same.")
        self.assertAlmostEqual(a.report()["ta"], theta, msg="After the first movement, the target azimuth should "
                                                            "still be the same.")
        self.assertAlmostEqual(a.report()["tr"], r - 1, msg="After the first movement, the target distance should be "
                                                            "the initial distance minus one.")
        self.assertEqual(a.report()["d"], 1, msg="After the first movement, the agent's distance travelled should be "
                                                 "one.")
        self.assertEqual(a.report()["st"], AgentStates.MOVING, msg="After the first movement, the agent's state "
                                                                   "should be MOVING")



import unittest
from sadie.agents.spatial import SpatialAgent
from random import randint
import numpy as np
from numpy import pi as π


class TestSpatialAgent(unittest.TestCase):
    def setUp(self) -> None:
        self.xinit, self.yinit = randint(-100, 100), randint(-100, 100)
        self.agent = SpatialAgent(x_init=self.xinit, y_init=self.yinit)

    def test_position(self):
        self.assertEqual(self.agent.position, (self.xinit, self.yinit),
                         msg="Agent does not return the correct initial position without operations.")

    def test_x(self):
        self.assertEqual(self.agent.x, self.agent.position[0])
        self.assertEqual(self.agent.x, self.xinit)

    def test_y(self):
        self.assertEqual(self.agent.y, self.agent.position[1])
        self.assertEqual(self.agent.y, self.yinit)

    def test_distance_from(self):
        agent = SpatialAgent(x_init=0, y_init=0)
        self.assertEqual(agent.distance_from(point=(1, 1)), np.sqrt(2))
        self.assertEqual(agent.distance_from(point=(0, 0)), 0)
        self.assertEqual(agent.distance_from(point=(-1, 0)), 1)
        self.assertEqual(agent.distance_from(point=(0, -1)), 1)
        self.assertEqual(agent.distance_from(point=(1, 0)), 1)
        self.assertEqual(agent.distance_from(point=(0, 1)), 1)

    def test_vector_to(self):
        agent = SpatialAgent(x_init=0, y_init=0)
        self.assertEqual(agent.vector_to((1, 0)), (0, 1))
        self.assertEqual(agent.vector_to((0, 1)), (π/2, 1))
        self.assertEqual(agent.vector_to((0, -1)), (-π/2, 1))
        self.assertEqual(agent.vector_to((-1, 0)), (π, 1))



from abc import ABC
from enum import Enum
from typing import Iterable, Type, List
import pandas as pd
from sadie.agents import AbstractAgent
from sadie.models.exceptions import ModelRunNotCompletedError


class ModelStatus(Enum):
    NOT_RUN = 0
    RUNNING = 1
    DONE = 2


class AbstractModel(ABC):
    """
    `AbstractModel` is an abstract base class for models that implements the fundamental functionalities of a model:

    * collecting and managing agents (through the `agents` property and the `populate()` method),
    * keeping track of time (through the `time` property), and
    * running the model (using the `run()` method) and extracting results from the collector to a pandas DataFrame (`to_df()` method.)

    It is a non-self-passing model, i.e. it does not pass any information about itself (or itself to begin with) to the
    agents it runs.
    """
    def __init__(self, init_time: int = 0, max_time: int = 100, *args, **kwargs):
        self.init_time: int = init_time
        self.max_time: int = max_time
        self._time: float = self.init_time
        self.time_step: float = kwargs.get("time_step", 1)
        self._collector: list = []
        self._agents: list = []
        self._state = ModelStatus.NOT_RUN

    def add_agent(self, *agents: AbstractAgent):
        """
        Adds one or more agents `agents` that implements the top-level abstract base class for agents (`AbstractAgent`)
        to the model's population.

        At this point in time, the `add_agent()` method is permissive - it does not check whether the agent(s) added
        implement(s) or descend(s) from `AbstractAgent`, it just adds them regardless. This may change, and result in
        a `TypeError` in the future.

        :param agents: agent(s) to be added
        """
        for agent in agents:
            self._agents.append(agent)

    @property
    def agents(self) -> List[AbstractAgent]:
        """
        Returns the list of agents currently within the model's population.

        :return: list of agents currently within the model's population.
        """
        return self._agents

    @agents.setter
    def agents(self, agents: Iterable):
        self._agents = agents

    @property
    def time(self) -> float:
        """
        Returns the current time of the model.

        :return: current model time
        """
        return self._time

    @time.setter
    def time(self, value):
        raise ValueError("Cannot set time directly.")

    @property
    def state(self) -> ModelStatus:
        """
        Returns the current state of the model.

        :return: current model state
        """
        return self._state

    @state.setter
    def state(self, value):
        raise ValueError("Cannot set status: model state is set directly by the model's own processes.")

    def populate(self, n_agents: int, agent_class: Type[AbstractAgent], *args, **kwargs):
        """
        Populates the model with `n_agents` agents of the class `agent_class`. `*args` and `**kwargs` are passed on to
        the constructor responsible for instantiating agents of the specified class.

        :param n_agents: number of agents to instantiate
        :param agent_class: class from which agents are instantiated
        """
        for i in range(n_agents):
            self.add_agent(agent_class(*args, **kwargs))

    def run(self):
        """
        Runs the model.
        """
        while self._time < self.max_time:
            self._state = ModelStatus.RUNNING
            for agent in self.agents:
                agent.update()
                report = agent.report()
                report["time"] = self._time
                self._collector.append(report)
            self._time += self.time_step
        self._state = ModelStatus.DONE

    def to_df(self) -> pd.DataFrame:
        """
        Exports the model's state as a `pandas` `DataFrame`.

        :return: model state as `pandas` `DataFrame`
        :raises ModelRunNotCompletedError: if the model is still in a running state.
        """
        if self._state == ModelStatus.DONE:
            return pd.DataFrame.from_dict(self._collector)
        else:
            raise ModelRunNotCompletedError
